/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  codewordEnumerator.cpp
///  \brief  Implements a gray code like run over all codewords.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   14.10.2010
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "codewordEnumerator.hpp"
#include "field.hpp"
#include "f2_f2.hpp"
#include "z4.hpp"
#include "matrix.hpp"

template<>
CodewordEnumerator<Field>::CodewordEnumerator(const std::vector<unsigned int>& type, const Matrix<Field>& mat) :
    m_type(type)
{
    m_n = mat.getCols();
    m_k = mat.getRows();
    m_actCodeword = CW(m_n, Element<Field>(0));

    m_rows = Field::get_r() * m_k;
    m_plus.resize(m_rows);
    m_minus.resize(m_rows);
    m_max.resize(m_rows + 1, Field::get_primefield_size());
    m_inf_word.resize(m_rows + 1, 0);
    m_direction.resize(m_rows + 1, 1);
    p_stopFlag = &m_inf_word.back();

    m_maxIterBeg = m_max.begin();
    m_inf_wordIterBeg = m_inf_word.begin();
    m_directionIterBeg = m_direction.begin();

    Matrix<Field> t = mat.transpose();
    const std::vector<Vec<Field> >& rows(t.getEntries());
    std::vector<Vec<Field> >::const_iterator rowIter(rows.begin());
    std::vector<Vec<Field> >::const_iterator rowIterEnd(rows.end());
    std::vector<CW>::iterator plusIter(m_plus.begin());
    std::vector<CW>::iterator minusIter(m_minus.begin());

    for (; rowIter != rowIterEnd; ++rowIter)
    {
        Element<Field> factor = Element<Field> (1);
        for (unsigned int exp = 0; exp < Field::get_r(); ++exp, factor *= Field::get_primitiveElement())
        {
            *plusIter = ((*rowIter) * factor);
            ++plusIter;
            *minusIter = ((*rowIter) * (-factor));
            ++minusIter;
        }
    }
}

template<>
Vec<Field> CodewordEnumerator<Field>::getInformationWord(void) const
{
    std::vector<int>::const_iterator infWordIter(m_inf_word.begin());
    Vec<Field> res;
    for(unsigned int i=0; i<m_k; ++i)
    {
        Element<Field> ent(0), factor(1);
        for(unsigned int i=0 ; i<Field::get_r() ; ++i, factor *= Field::get_primitiveElement(), ++infWordIter)
        {
            ent += factor*Field::primeFieldElement(*infWordIter);
        }
        res.push_back(ent);
    }
    return res;
}

template<>
Vec<BinaryField> CodewordEnumerator<BinaryField>::getInformationWord(void) const
{
    Vec<BinaryField> res;
    std::vector<int>::const_iterator infWordIter(m_inf_word.begin());
    for(unsigned int i=0; i<m_k; ++i, ++infWordIter)
    {
        res.push_back(*infWordIter);
    }
    return res;
}

template<>
Vec<Z4> CodewordEnumerator<Z4>::getInformationWord(void) const
{
    Vec<Z4> res;
    std::vector<int>::const_iterator infWordIter(m_inf_word.begin());
    for(unsigned int i=0; i<m_k; ++i, ++infWordIter)
    {
        res.push_back(*infWordIter);
    }
    return res;
}

template<>
Vec<F2_F2> CodewordEnumerator<F2_F2>::getInformationWord(void) const
{
    Vec<F2_F2> res;
    unsigned int k0(m_type[0]);
    unsigned int k1(m_type[1]);

    std::vector<int>::const_iterator infWordIter(m_inf_word.begin());
    for(unsigned int i=0; i<k0; ++i, ++infWordIter)
    {
        Element<F2_F2> elt(*infWordIter);
        ++infWordIter;
        elt += Element<F2_F2>(2*(*infWordIter));
        res.push_back(elt);
    }

    for(unsigned int i=0; i<k1; ++i, ++infWordIter)
    {
        res.push_back( *infWordIter );
    }
    return res;
}

//template<>
//uchar CodewordEnumerator<BinaryField>::getInformationWord(void) const
//{
//    uchar res(0);
//    unsigned int counter = 0;
//    for(std::vector<int>::iterator iter(m_inf_wordIterBeg); iter<m_inf_wordIterBeg+m_rows; ++iter, ++counter)
//    {
//        if(*iter)
//        {
//            res += ((uchar) 1)<<counter;
//        }
//    }
//    return res;
//}

#if not F2M_CASE
template<>
CodewordEnumerator<BinaryField>::CodewordEnumerator(const std::vector<unsigned int>& type,
        const Matrix<BinaryField>& mat) :
    m_type(type)
{
    m_n = mat.getCols();
    m_k = mat.getRows();
    m_actCodeword = CW(m_n);

    m_rows = m_k;
    m_plus.resize(m_rows);
    m_minus.resize(m_rows);
    m_max.resize(m_rows + 1, 2);
    m_inf_word.resize(m_rows + 1, 0);
    m_direction.resize(m_rows + 1, 1);
    p_stopFlag = &m_inf_word.back();

    m_maxIterBeg = m_max.begin();
    m_inf_wordIterBeg = m_inf_word.begin();
    m_directionIterBeg = m_direction.begin();

    Matrix<BinaryField> t = mat.transpose();
    const std::vector<Vec<BinaryField> >& rows(t.getEntries());
    std::vector<Vec<BinaryField> >::const_iterator rowIter(rows.begin());
    std::vector<Vec<BinaryField> >::const_iterator rowIterEnd(rows.end());
    std::vector<CW>::iterator plusIter(m_plus.begin());
    std::vector<CW>::iterator minusIter(m_minus.begin());

    for (; rowIter != rowIterEnd; ++rowIter)
    {
        *plusIter = *rowIter;
        ++plusIter;
        *minusIter = *rowIter;
        ++minusIter;
    }
}

template<>
CodewordEnumerator<Z4>::CodewordEnumerator(const std::vector<unsigned int>& type, const Matrix<Z4>& matrix) :
    m_type(type)
{
    Matrix<Z4> mat(matrix);
    mat.toOrderedBasis(false, NULL, NULL);
    m_n = mat.getCols();
    m_k = mat.getRows();
    m_actCodeword = CW(m_n);

    m_rows = type[0] + type[1];
    m_plus.resize(m_rows);
    m_minus.resize(m_rows);
    m_max.resize(m_rows + 1, 4);
    for (unsigned int i = type[0]; i <= m_rows; ++i)
    {
        m_max[i] = 2;
    }

    m_inf_word.resize(m_rows + 1, 0);
    m_direction.resize(m_rows + 1, 1);
    p_stopFlag = &m_inf_word.back();

    m_maxIterBeg = m_max.begin();
    m_inf_wordIterBeg = m_inf_word.begin();
    m_directionIterBeg = m_direction.begin();

    Matrix<Z4> t = mat.transpose();
    const std::vector<Vec<Z4> >& rows(t.getEntries());
    std::vector<Vec<Z4> >::const_iterator rowIter(rows.begin());
    std::vector<Vec<Z4> >::const_iterator rowIterEnd(rows.end());
    std::vector<DenseVec<Z4> >::iterator plusIter(m_plus.begin());
    std::vector<DenseVec<Z4> >::iterator minusIter(m_minus.begin());

    Element<Z4> minusOne = -Element<Z4> (1);

    for (; rowIter != rowIterEnd; ++rowIter)
    {
        *plusIter = *rowIter;
        ++plusIter;
        *minusIter = (*rowIter) * minusOne;
        ++minusIter;
    }
}


template<>
CodewordEnumerator<F2_F2>::CodewordEnumerator(const std::vector<unsigned int>& type, const Matrix<F2_F2>& matrix) :
    m_type(type)
{
    Matrix<F2_F2> mat(matrix);
    mat.toOrderedBasis(false, NULL, NULL);
    m_n = mat.getCols();
    m_k = mat.getRows();
    m_actCodeword = CW(m_n);

    m_rows = 2*type[0] + type[1];
    m_plus.resize(m_rows);
    m_minus.resize(m_rows);
    m_max.resize(m_rows + 1, 2);
    m_inf_word.resize(m_rows + 1, 0);
    m_direction.resize(m_rows + 1, 1);
    p_stopFlag = &m_inf_word.back();

    m_maxIterBeg = m_max.begin();
    m_inf_wordIterBeg = m_inf_word.begin();
    m_directionIterBeg = m_direction.begin();

    Matrix<F2_F2> t = mat.transpose();
    const std::vector<Vec<F2_F2> >& rows(t.getEntries());
    std::vector<Vec<F2_F2> >::const_iterator rowIter(rows.begin());
    std::vector<Vec<F2_F2> >::const_iterator rowIterEnd(rows.begin() + type[0]);
    std::vector<DenseVec<F2_F2> >::iterator plusIter(m_plus.begin());
    std::vector<DenseVec<F2_F2> >::iterator minusIter(m_minus.begin());

    Element<F2_F2> X(2);

    for (; rowIter != rowIterEnd; ++rowIter)
    {
        *plusIter = *rowIter;
        *minusIter = *plusIter;
        ++plusIter;
        ++minusIter;

        *plusIter = *rowIter;
        *plusIter *= X;
        *minusIter = *plusIter;
        ++plusIter;
        ++minusIter;
    }

    rowIterEnd = rows.end();
    for (; rowIter != rowIterEnd; ++rowIter)
    {
        *plusIter = *rowIter;
        *minusIter = *plusIter;
        ++plusIter;
        ++minusIter;
    }
}

template<>
bool CodewordEnumerator<BinaryField>::isProjPointRepresentative(void) const
{
    return true;
}


template<>
bool CodewordEnumerator<Z4>::isProjPointRepresentative(void) const
{
    std::vector<int>::const_iterator inf_wordIter = m_inf_word.begin();
    std::vector<int>::const_iterator inf_wordIterEnd = inf_wordIter + m_type.front();
    for (; inf_wordIter < inf_wordIterEnd; ++inf_wordIter)
    {
        if (*inf_wordIter == 1)
        {
            return true;
        }
        if (*inf_wordIter == 3)
        {
            return false;
        }
    }
    inf_wordIter = inf_wordIterEnd;
    inf_wordIterEnd = m_inf_word.end();
    for (; inf_wordIter < inf_wordIterEnd; ++inf_wordIter)
    {
        if (*inf_wordIter == 1)
        {
            return true;
        }
    }
    return false;
}

template<>
bool CodewordEnumerator<F2_F2>::isProjPointRepresentative(void) const
{
    std::vector<int>::const_iterator inf_wordIter = m_inf_word.begin();
    std::vector<int>::const_iterator inf_wordIterEnd = inf_wordIter + 2*m_type[0];
    for (; inf_wordIter < inf_wordIterEnd; ++inf_wordIter)
    {
        if (*inf_wordIter == 1)
        {
            ++inf_wordIter;
            if (*inf_wordIter == 1)
            {
                return false; // the leading coefficient is equal to 3
            }
            return true;// the leading coefficient is equal to 1
        }
        else
        {
            ++inf_wordIter;
        }
    }
    // the information word does not contain a unit in its upper part, there must be at least one in the lower for returning true
    inf_wordIterEnd = m_inf_word.end();
    for (; inf_wordIter < inf_wordIterEnd; ++inf_wordIter)
    {
        if (*inf_wordIter == 1)
        {
            return true;
        }
    }
    return false;
}

#endif


#if F2M_CASE

template<>
bool CodewordEnumerator<Field>::isProjPointRepresentative(void) const
{
    return m_actCodeword.isProjPointRepresentative();
}

#else
template<>
bool CodewordEnumerator<Field>::isProjPointRepresentative(void) const
{
    DenseVec<Field>::const_reverse_iterator actCodewordIterEnd(m_actCodeword.rend());
    for (DenseVec<Field>::const_reverse_iterator actCodewordIter(m_actCodeword.rbegin()); actCodewordIter < actCodewordIterEnd; ++actCodewordIter)
    {
        if (!actCodewordIter->isZero())
        {
            if (actCodewordIter->isOne())
            {
                return true;
            }
            else
            {
                return false;
            }
        }
    }
    return false; // the zero codeword
}
#endif
