/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  magmaCASModule.h
///  \brief Provides an interface to MAGMA.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef MAGMACASMODULE_H_
#define MAGMACASMODULE_H_

#include "abstractCASModule.h"
#include "field.hpp"
//#include "weights.hpp"

template<typename RING, AlgTypeE ALGTYPE>
class ActingElement;

class MagmaCASModule : private AbstractCASModule
{
public:
	MagmaCASModule(char* call[], const std::string& packagesFolder);
	~MagmaCASModule();

	template<typename RING>
	void initRingStructure(void);

	void initTrivialSn(const std::string& group, int n);
	void getFather(const std::string& group, std::vector<int>& father, My_Long_Int& group_order );
	void extendGroup(const std::string& group, const Permutation& newgen, int n) ;
	void buildGroup(const std::string& group, const std::vector<Permutation>& newgens, int n) ;
	My_Long_Int getOrder( const std::string& group );
	unsigned int BKLCUpperBound( unsigned int n, unsigned int k, unsigned int q);
	unsigned int BKLCLowerBound( unsigned int n, unsigned int k, unsigned int q);

	template<typename RING>
	bool verifyMinDist(const Matrix<RING>& matrix, unsigned int d, unsigned int dual_d);

    template<typename RING>
    unsigned int  getMinDist(const Matrix<RING>& matrix);

	template<typename RING>
	bool NebeCheck( const Matrix<RING>& m, unsigned int desired_d);

    void fillGLsubgroup( unsigned int k, const std::vector<ActingElement<Field, PLUSMINUSMONOMIAL> >& autom_gens, const std::string& groupname);
    void calcStabilizer(const std::string& groupname, const std::string& stabname, const Vec<Field>& v);

    template<typename RING>
    void calcOrbitReps(unsigned int k, const std::string& groupname,  std::vector<Vec<RING > >& orbitReps );

    template<typename RING>
    void getMinWords(const Matrix<RING>matrix, std::vector<DenseVec<RING> >& res);

//	static bool checkBKLCMaximum(unsigned int n, unsigned int q)


//	template<unsigned int chainlength>
//	unsigned int dualMinDist(const unsigned int* const type, unsigned int n, unsigned int q,
//	        const Distribution<SymWeight<chainlength> >& primal_wd);
//
//	unsigned long getMinWords(const Matrix<Z4>& matrix, vector<Vec<Z4> >& minWords);
private:
    const std::string m_ringname, m_toElt, m_toInt;
	bool verifyMinDist(unsigned int d, unsigned int dual_d);

	  static bool checkBKLCMaximum(unsigned int n, unsigned int q)
	    {
	        switch(q)
	        {
	        case 2:
	        case 4: return n<=256;
	        case 3: return n<=243;
	        case 5:
	        case 8:
	        case 9: return n<=130;
	        case 7: return n<=100;
	        default: return false;
	        }
	    }
};

template<typename RING>
bool MagmaCASModule::NebeCheck( const Matrix<RING>& m, unsigned int desired_d)
{
    m_to<<"Gamma := Matrix(F, "<<m.getRows()<<","<<m.getCols()<<", [";
    m.print(m_to, ",", true, "t");
    m_to<<"]);";

    m_to<<"C := getCode(Gamma);"<<std::endl;
    m_to<<"X := VerifyMinimumDistanceUpperBound(C, "<< desired_d-1 <<" );"<<std::endl;
    m_to<<"lb, ub := MinimumWeightBounds(C);"<<std::endl;
    m_to<<"print ub;"<<std::endl;
    unsigned int ub;
    m_from>>ub;
    //std::cout<<ub<<std::endl;
    return !(ub < desired_d );
}

template<>
bool MagmaCASModule::verifyMinDist(const Matrix<Field>& matrix, unsigned int d, unsigned int dual_d);

template<>
bool MagmaCASModule::verifyMinDist(const Matrix<BinaryField>& matrix, unsigned int d, unsigned int dual_d);

template<typename RING>
bool MagmaCASModule::verifyMinDist(const Matrix<RING>& matrix, unsigned int d, unsigned int dual_d)
{
    return false;
}

template<typename RING>
unsigned int MagmaCASModule::getMinDist(const Matrix<RING>& matrix)
{
    m_to<<"MinimumDistance(LinearCode(Matrix("<<m_ringname<<", "<<matrix.getRows()<<","<<matrix.getCols()<<", [";
        matrix.print(m_to, ",", &m_toElt);
        m_to<<"])));"<<std::endl;
    unsigned int d;
    m_from>>d;
    return d;
}

template<typename RING>
void MagmaCASModule::getMinWords(const Matrix<RING>matrix, std::vector<DenseVec<RING> >& res)
{
    res.clear();
    m_to<<"C := LinearCode(Matrix("<<m_ringname<<"," << matrix.getRows()<<", "<<matrix.getCols()<< ", [";
    matrix.print(m_to, ",", &m_toElt);
    m_to<<"]));"<<std::endl;

    unsigned int size;
    m_to<<"words := {Normalize(w): w in MinimumWords(C) };"<<std::endl;
    m_to<<"old_size := 0;"<<std::endl;
    m_to<<"next_w := MinimumDistance(C);"<<std::endl;
    m_to<<"while #words eq old_size or Dimension(sub<C|words>) ne Dimension(C) do"<<std::endl;
    m_to<<"old_size := #words;"<<std::endl;
    m_to<<"next_w +:= 1;"<<std::endl;
    m_to<<"words join:= {Normalize(w): w in Words(C, next_w) };"<<std::endl;
    m_to<<"end while;"<<std::endl;
    m_to<<"print #words;"<<std::endl;
    m_from>>size;

    m_to<<"for c in words do"<<std::endl;
    m_to<<"s :=  ElementToSequence(c);"<<std::endl;
    m_to<<"for x in s do"<<std::endl;
    m_to<<"print "<<m_toInt<<"(x);"<<std::endl;
    m_to<<"end for;"<<std::endl;
    m_to<<"end for;"<<std::endl;

    Vec<RING> c(matrix.getCols());
    typename Vec<RING>::iterator cIterBeg(c.begin());
    typename Vec<RING>::iterator cIterEnd(c.end());

    res = std::vector<DenseVec<RING> >(size, DenseVec<RING>(matrix.getCols()));
    typename std::vector<DenseVec<RING> >::iterator resIter(res.begin());
    typename std::vector<DenseVec<RING> >::iterator resIterEnd(res.end());
    unsigned int dummy;
    for(; resIter<resIterEnd; ++resIter)
    {
        typename Vec<RING>::iterator cIter( cIterBeg );
        for( ; cIter<cIterEnd; ++cIter)
        {
            m_from>>dummy;
            *cIter = Element<RING>(dummy);
        }
        *resIter = c;
    }
}



//template<>
//unsigned int dualMinDist<1u>(const unsigned int* const type, unsigned int n, const Distribution<SymWeight<1u> >& primal_wd);
//
//template<>
//unsigned int dualMinDist<2u>(const unsigned int* const type, unsigned int n, const Distribution<SymWeight<2u> >& primal_wd);

#endif /*MAGMACASMODULE_H_*/
